import asyncio
from collections.abc import AsyncIterator
from dataclasses import dataclass
from datetime import datetime
from typing import cast

from fastapi import FastAPI
from fastapi.concurrency import asynccontextmanager

from fluid.scheduler import TaskInfo
from fluid.scheduler.broker import RedisTaskBroker
from fluid.scheduler.consumer import TaskManager
from fluid.tools_fastapi import backdoor
from fluid.utils.http_client import HttpxClient


@asynccontextmanager
async def start_fastapi(app: FastAPI) -> AsyncIterator:
    backdoor.setup(app, port=0)
    async with app.router.lifespan_context(app):
        yield app


def redis_broker(task_manager: TaskManager) -> RedisTaskBroker:
    return cast(RedisTaskBroker, task_manager.broker)


@dataclass
class TaskClient(HttpxClient):
    url: str = "http://test_api"

    async def get_tasks(self) -> list[TaskInfo]:
        data = await self.get(f"{self.url}/tasks")
        return [TaskInfo(**task) for task in data]

    async def get_task(self, task_name: str) -> TaskInfo:
        data = await self.get(f"{self.url}/tasks/{task_name}")
        return TaskInfo(**data)

    async def wait_for_task(
        self,
        task_name: str,
        last_run_end: datetime | None = None,
        timeout: float = 1.0,
    ) -> TaskInfo:
        sleep = min(timeout / 10.0, 0.1)
        async with asyncio.timeout(timeout):
            while True:
                task = await self.get_task(task_name)
                if task.last_run_end != last_run_end:
                    return task
                await asyncio.sleep(sleep)



import asyncio

from fluid.utils.worker import Worker


class SimpleWorker(Worker):
    async def run(self):
        while self.is_running():
            self.print_message()
            await asyncio.sleep(1)

    def print_message(self):
        print(f"Hello from {self.worker_name} in state {self.worker_state}")


async def main():
    worker = SimpleWorker()
    worker.print_message()
    await worker.startup()
    asyncio.get_event_loop().call_later(5, worker.gracefully_stop)
    await worker.wait_for_shutdown()
    worker.print_message()


if __name__ == "__main__":
    asyncio.run(main())



from __future__ import annotations

from contextlib import asynccontextmanager
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, AsyncIterator, Self

import sqlalchemy as sa
from sqlalchemy.engine import Engine, create_engine
from sqlalchemy.engine.url import make_url
from sqlalchemy.ext.asyncio import AsyncConnection, AsyncEngine, create_async_engine

from fluid import settings

from .migration import Migration


@dataclass
class Database:
    """A container for tables in a database and a manager of asynchronous
    connections to a postgresql database
    """

    dsn: str
    """data source name, aka connection string

    Example: `postgresql+asyncpg://user:password@localhost/dbname`

    Note that the `+asyncpg` part is important for the async engine.

    Currently, only `postgresql+asyncpg` is supported, but other databases may
    be supported in the future.
    """
    echo: bool = settings.DBECHO
    """Echo SQL queries to stdout

    It defaults to the `DBECHO` setting in the settings module
    """
    pool_size: int = settings.DBPOOL_MAX_SIZE
    max_overflow: int = settings.DBPOOL_MAX_OVERFLOW
    metadata: sa.MetaData = field(default_factory=sa.MetaData)
    migration_path: str | Path = ""
    """Path to the directory containing migration files. If empty, migrations will
    be stored in the default location `migrations` in the current working directory.
    """
    app_name: str = settings.APP_NAME
    _engine: AsyncEngine | None = None

    @classmethod
    def from_env(
        cls,
        *,
        dsn: str = settings.DATABASE,
        schema: str | None = settings.DATABASE_SCHEMA,
        **kwargs: Any,
    ) -> Self:
        """Create a new database container from environment variables as defaults"""
        return cls(dsn=dsn, metadata=sa.MetaData(schema=schema), **kwargs)

    @property
    def tables(self) -> dict[str, sa.Table]:
        """A dictionary of tables in the database"""
        return self.metadata.tables

    @property
    def engine(self) -> AsyncEngine:
        """The :class:`sqlalchemy.ext.asyncio.AsyncEngine` creating connection
        and transactions"""
        if self._engine is None:
            self._engine = create_async_engine(
                self.dsn,
                echo=self.echo,
                pool_size=self.pool_size,
                max_overflow=self.max_overflow,
                connect_args=dict(server_settings=dict(application_name=self.app_name)),
            )
        return self._engine

    @property
    def url(self) -> sa.engine.URL:
        """The SQLAlchemy URL object for the database"""
        return make_url(self.dsn)

    @property
    def sync_engine(self) -> Engine:
        """The sqlalchemy Engine object for synchrouns operations"""
        url = self.url
        query = dict(url.query)
        if sslmode := query.pop("ssl", None):
            query["sslmode"] = sslmode
        url = url.set(drivername="postgresql", query=query)
        return create_engine(url)

    @asynccontextmanager
    async def connection(self) -> AsyncIterator[AsyncConnection]:
        """Context manager for obtaining an asynchronous connection"""
        async with self.engine.connect() as conn:
            yield conn

    @asynccontextmanager
    async def ensure_connection(
        self,
        conn: AsyncConnection | None = None,
    ) -> AsyncIterator[AsyncConnection]:
        """Context manager for obtaining an asynchronous connection"""
        if conn:
            yield conn
        else:
            async with self.engine.connect() as conn:
                yield conn

    @asynccontextmanager
    async def transaction(self) -> AsyncIterator[AsyncConnection]:
        """Context manager for initializing an asynchronous database transaction"""
        async with self.engine.begin() as conn:
            yield conn

    @asynccontextmanager
    async def ensure_transaction(
        self,
        conn: AsyncConnection | None = None,
    ) -> AsyncIterator[AsyncConnection]:
        """Context manager for ensuring we a connection has initialized
        a database transaction"""
        if conn:
            if not conn.in_transaction():
                async with conn.begin():
                    yield conn
            else:
                yield conn
        else:
            async with self.transaction() as conn:
                yield conn

    async def close(self) -> None:
        """Close the asynchronous db engine if opened"""
        if self._engine is not None:
            engine, self._engine = self._engine, None
            await engine.dispose()

    async def ping(self) -> str:
        """Ping the database"""
        # TODO: we need a custom ping query
        async with self.connection() as conn:
            await conn.execute(sa.text("SELECT 1"))
        return "ok"

    def migration(self) -> Migration:
        """The migration manager for this database"""
        return Migration(db=self)



import enum

import sqlalchemy as sa

from fluid.utils.dates import utcnow
from fluid.utils.text import create_uid


class TaskType(enum.Enum):
    todo = 0
    issue = 1


def meta(meta: sa.MetaData) -> None:
    """Add task related tables"""

    sa.Table(
        "tasks",
        meta,
        sa.Column(
            "id", sa.String(32), primary_key=True, doc="Unique ID", default=create_uid
        ),
        sa.Column(
            "title",
            sa.String(64),
            nullable=False,
            info=dict(min_length=3),
        ),
        sa.Column("created", sa.DateTime(timezone=True), default=utcnow),
        sa.Column("done", sa.DateTime(timezone=True)),
        sa.Column("severity", sa.Integer),
        sa.Column("created_by", sa.String, default="", nullable=False),
        sa.Column("type", sa.Enum(TaskType)),
        sa.Column("unique_title", sa.String, unique=True),
        sa.Column("story_points", sa.Numeric),
        sa.Column("random", sa.String(64)),
        sa.Column(
            "subtitle",
            sa.String(64),
            nullable=False,
            default="",
        ),
    )



import pytest

from fluid.scheduler.models import TaskInfo, TaskState
from fluid.utils.http_client import HttpResponseError
from fluid.utils.waiter import wait_for
from tests.scheduler.tasks import TaskClient

pytestmark = pytest.mark.asyncio(loop_scope="module")


async def test_get_tasks(cli: TaskClient) -> None:
    data = await cli.get(f"{cli.url}/tasks")
    assert len(data) == 13
    tasks = {task["name"]: TaskInfo(**task) for task in data}
    dummy = tasks["dummy"]
    assert dummy.name == "dummy"
    assert set(dummy.tags) == {"test", "slow"}


async def test_get_tasks_by_tags(cli: TaskClient) -> None:
    data = await cli.get(f"{cli.url}/tasks?tags=test")
    names = {task["name"] for task in data}
    assert names == {"fast", "dummy"}
    data = await cli.get(f"{cli.url}/tasks?tags=skip_db")
    names = {task["name"] for task in data}
    assert names == {"ping"}
    data = await cli.get(f"{cli.url}/tasks?tags=skip_db&tags=fast")
    names = {task["name"] for task in data}
    assert names == {"ping", "fast"}
    data = await cli.get(f"{cli.url}/tasks?tags=does-not-exist")
    assert data == []


async def test_get_tasks_status(cli: TaskClient) -> None:
    async def check() -> bool:
        data = await cli.get(f"{cli.url}/tasks-status")
        return bool(data.get("workers"))

    await wait_for(check, timeout=3.0)
    data = await cli.get(f"{cli.url}/tasks-status")
    assert len(data["workers"]) == 1
    assert data["workers"][0]["kind"] == "task_scheduler"
    assert data["workers"][0]["status"]
    assert "queues" in data


async def test_get_task_404(cli: TaskClient) -> None:
    with pytest.raises(HttpResponseError):
        await cli.get(f"{cli.url}/tasks/whatever")


async def test_run_task_404(cli: TaskClient) -> None:
    with pytest.raises(HttpResponseError):
        await cli.post(f"{cli.url}/tasks/whatever")


async def test_run_task(cli: TaskClient) -> None:
    task = await cli.get_task("dummy")
    assert task.last_run_end is None
    task = await cli.get_task("dummy")
    data = await cli.post(f"{cli.url}/tasks/dummy", json=dict(sleep=0.2))
    assert data["task"] == "dummy"
    # wait for task
    task = await cli.wait_for_task("dummy")
    assert task.last_run_end is not None


async def test_patch_task_404(cli: TaskClient) -> None:
    with pytest.raises(HttpResponseError):
        await cli.patch(f"{cli.url}/tasks/whatever", json=dict(enabled=False))


async def test_patch_task(cli: TaskClient) -> None:
    data = await cli.patch(f"{cli.url}/tasks/dummy", json=dict(enabled=False))
    assert data["enabled"] is False
    task = await cli.get_task("dummy")
    assert task.enabled is False
    data = await cli.post(f"{cli.url}/tasks/dummy", json=dict(sleep=0.2))
    task = await cli.wait_for_task("dummy", last_run_end=task.last_run_end)
    assert task.enabled is False
    assert task.last_run_state == TaskState.aborted
    assert task.last_run_end is not None


async def test_ping(cli: TaskClient) -> None:
    data = await cli.post(f"{cli.url}/tasks/ping")
    assert data["task"] == "ping"


async def test_delete_queue(cli: TaskClient) -> None:
    data = await cli.delete(f"{cli.url}/tasks-queue")
    assert set(data.keys()) == {"high", "medium", "low"}
    assert all(isinstance(v, int) for v in data.values())


async def test_delete_queue_by_priority(cli: TaskClient) -> None:
    data = await cli.delete(f"{cli.url}/tasks-queue/high")
    assert set(data.keys()) == {"high"}
    assert isinstance(data["high"], int)



import asyncio
import json
import os
from dataclasses import dataclass

import pytest
from redis.asyncio import Redis

from fluid.scheduler import (
    TaskPriority,
    TaskRun,
    TaskScheduler,
    TaskState,
)
from fluid.scheduler.errors import UnknownTaskError
from fluid.utils.stacksampler import Sampler
from fluid.utils.waiter import wait_for

pytestmark = pytest.mark.asyncio(loop_scope="module")


async def test_scheduler_manager(task_scheduler: TaskScheduler) -> None:
    assert task_scheduler
    assert task_scheduler.broker.registry
    assert task_scheduler.type == "task_scheduler"
    assert "dummy" in task_scheduler.registry
    assert "ping" in task_scheduler.registry
    with pytest.raises(UnknownTaskError):
        task_scheduler.broker.task_from_registry("bbdjchbjch")


async def test_queue_length(task_scheduler: TaskScheduler) -> None:
    ql = await task_scheduler.broker.queue_length()
    assert len(ql) == 3
    for p in TaskPriority:
        assert ql[p.name] >= 0


async def test_execute(task_scheduler: TaskScheduler) -> None:
    task_run = await task_scheduler.execute("dummy")
    assert task_run.name == "dummy"
    assert task_run.end


async def test_dummy_queue(task_scheduler: TaskScheduler) -> None:
    task_run = await task_scheduler.execute("dummy", sleep=0.2)
    assert task_run.is_done


async def test_dummy_error(task_scheduler: TaskScheduler) -> None:
    with pytest.raises(RuntimeError):
        await task_scheduler.execute("dummy", error=True)


async def test_dummy_abort(task_scheduler: TaskScheduler) -> None:
    task_run = await task_scheduler.queue_and_wait("dummy", abort=True)
    assert task_run.state == TaskState.aborted


async def test_dummy_rate_limit(task_scheduler: TaskScheduler) -> None:
    s1, s2 = await asyncio.gather(
        task_scheduler.queue_and_wait("dummy", sleep=2, timeout=10),
        task_scheduler.queue_and_wait("dummy", sleep=1, timeout=10),
    )
    current = await task_scheduler.broker.current_task_runs("dummy")
    assert current == 0
    assert s1.is_done
    assert s2.is_done
    assert s1.state is TaskState.rate_limited or s2.state is TaskState.rate_limited


async def test_cpu_bound_execution(
    task_scheduler: TaskScheduler, redis: Redis  # type: ignore
) -> None:
    task_run = await task_scheduler.queue_and_wait("cpu_bound", sleep=1.0, timeout=5)
    assert task_run.end
    result = await redis.get(task_run.id)
    assert result
    data = json.loads(result)
    assert data["pid"] != os.getpid()
    assert data["sleep"] == 1.0


async def test_cpu_bound_env(
    task_scheduler: TaskScheduler, redis: Redis  # type: ignore
) -> None:
    task_run = await task_scheduler.queue_and_wait("cpu_bound_env", timeout=5)
    assert task_run.end
    result = await redis.get(task_run.id)
    assert result
    assert result.decode() == "hello_from_env"


async def test_cpu_bound_failure(task_scheduler: TaskScheduler) -> None:
    task_run = await task_scheduler.queue_and_wait("cpu_bound", error=True, timeout=5)
    assert task_run.state == TaskState.failure


async def test_cpu_bound_abort(task_scheduler: TaskScheduler) -> None:
    task_run = await task_scheduler.queue_and_wait("cpu_bound", abort=True, timeout=5)
    assert task_run.state == TaskState.aborted


async def test_task_info(task_scheduler: TaskScheduler) -> None:
    with pytest.raises(UnknownTaskError):
        await task_scheduler.broker.enable_task("hfgfhgfhfh")
    info = await task_scheduler.broker.enable_task("ping")
    assert info.enabled is True
    assert info.name == "ping"
    assert info.schedule == "every(0:00:02)"
    assert info.description == "A simple scheduled task that ping the broker"


async def test_disabled_execution(task_scheduler: TaskScheduler) -> None:
    info = await task_scheduler.broker.enable_task("add", enable=False)
    assert info.enabled is False
    assert info.name == "add"
    task_run = await task_scheduler.queue_and_wait(
        "add",
        priority=TaskPriority.high,
        a=3,
        b=4,
    )
    assert task_run.name == "add"
    assert task_run.params.a == 3
    assert task_run.params.b == 4
    assert task_run.end
    assert task_run.state == TaskState.aborted.name


async def test_timeout(task_scheduler: TaskScheduler) -> None:
    task_run = await task_scheduler.queue_and_wait("fast", sleep=5, timeout=5)
    assert task_run.state == TaskState.failure
    assert task_run.end


@dataclass
class AsyncHandler:
    task_run: TaskRun | None = None

    async def __call__(self, task_run: TaskRun) -> None:
        await asyncio.sleep(0.1)
        self.task_run = task_run


async def test_async_handler(task_scheduler: TaskScheduler) -> None:
    handler = AsyncHandler()
    task_scheduler.register_async_handler("running.test", handler)
    task_run = await task_scheduler.queue_and_wait("dummy")
    assert task_run.state == TaskState.success
    await wait_for(lambda: handler.task_run is not None)
    assert handler.task_run
    assert handler.task_run.state == TaskState.running
    assert task_scheduler.unregister_async_handler("running.test") is handler


async def test_sampler(sampler: Sampler) -> None:
    assert sampler.started
    stats = sampler.stats()
    assert stats
    async with sampler.flamegraph_file("test", stats) as data:
        assert data



from fluid.utils import kernel
from tests.scripts import long_line


async def test_env_variable() -> None:
    result = kernel.CollectBytes()
    await kernel.run("env", result_callback=result, env=dict(KERNEL_TEST_ME="yes"))
    assert result.data == b"KERNEL_TEST_ME=yes\n"


async def test_long_line() -> None:
    result = kernel.CollectBytes()
    code = await kernel.run(
        "python",
        "-m",
        "tests.scripts.long_line",
        result_callback=result,
        stream_output=True,
        stream_error=True,
    )
    assert code == 0
    assert result.data
    lines = [text for text in result.data.decode("utf-8").split("\n") if text]
    assert len(lines) == 1
    assert len(lines[0]) == long_line.length


async def test_long_lines() -> None:
    result = kernel.CollectBytes()
    code = await kernel.run(
        "python",
        "-m",
        "tests.scripts.long_line",
        "3",
        result_callback=result,
        stream_output=True,
        stream_error=True,
    )
    assert code == 0
    assert result.data
    lines = [text for text in result.data.decode("utf-8").split("\n") if text]
    assert len(lines) == 3
    for text in lines:
        assert len(text) == long_line.length



from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING

from kubernetes_asyncio import client, config
from kubernetes_asyncio.client.api_client import ApiClient
from slugify import slugify

from .common import cpu_env
from .errors import TaskAbortedError, TaskRunError

if TYPE_CHECKING:
    from .models import K8sConfig, TaskRun


def get_job_name(ctx: TaskRun) -> str:
    short_task_id = ctx.id[:7]
    job_name = slugify(f"task-{ctx.name}-{short_task_id}")[:63]
    return job_name


async def run_on_k8s_job(ctx: TaskRun) -> None:
    """Run a task on a k8s job

    This is available when running inside a Kubernetes cluster.
    Only task consumer/scheduler with command line client can use this
    """
    task = ctx.task
    job_name = get_job_name(ctx)
    # load k8s config from within the cluster
    config.load_incluster_config()
    k8s_config = task.get_k8s_config()
    async with ApiClient() as api:
        v1 = client.AppsV1Api(api)
        # get the targeted deployment from the targeted namespace
        tasks = await v1.read_namespaced_deployment(
            k8s_config.deployment,
            k8s_config.namespace,
        )
        # pod template
        pod_template = k8s_job_pod_template(
            ctx,
            tasks.spec.template,
            k8s_config,
        )
        batch = client.BatchV1Api(api)
        job = client.V1Job(
            metadata=client.V1ObjectMeta(name=job_name),
            spec=client.V1JobSpec(
                ttl_seconds_after_finished=k8s_config.job_ttl,
                backoff_limit=0,
                template=pod_template,
            ),
        )
        response = await batch.create_namespaced_job(k8s_config.namespace, job)
        ctx.logger.info(f"Job created. status={response.status}")
        while True:
            job_status = await batch.read_namespaced_job_status(
                name=job_name,
                namespace=k8s_config.namespace,
            )
            if job_status.status.succeeded:
                ctx.logger.info(f"status={job_status.status}")
                if reason := await ctx.task_manager.broker.get_task_aborted(ctx.id):
                    raise TaskAbortedError(reason)
                break
            if job_status.status.failed is not None:
                raise TaskRunError(f"K8s task failed status {job_status.status}")
            await asyncio.sleep(k8s_config.sleep)


def k8s_job_pod_template(
    ctx: TaskRun, pod_template: client.V1PodTemplateSpec, k8s_config: K8sConfig
) -> client.V1PodTemplateSpec:
    """Modify the pod template for the k8s job."""
    # pod spec for the job is based on the deployment's pod spec,
    # but with a different command and args
    pod_spec = pod_template.spec
    # get the targeted container from the pod spec
    container = next(
        (c for c in pod_spec.containers if c.name == k8s_config.container),
        None,
    )
    if container is None:
        raise TaskRunError(f"Container {k8s_config.container} not found")
    command = list(container.command or [])
    if command and command[-1] == "serve":
        command.pop()
    container.command = command
    container.args = [
        "exec",
        ctx.name,
        "--log",
        "--run-id",
        ctx.id,
        "--params",
        ctx.params.model_dump_json(),
    ]
    # resources
    if resources := k8s_config.resources:
        container.resources = client.V1ResourceRequirements(**resources)
    # env vars for the task context
    env = list(container.env or [])
    for name, value in cpu_env().items():
        env.append(client.V1EnvVar(name=name, value=value))
    for name, value in ctx.task.env.items():
        env.append(client.V1EnvVar(name=name, value=value))
    container.env = env
    container.liveness_probe = None  # type: ignore[assignment]
    container.readiness_probe = None  # type: ignore[assignment]
    pod_spec.containers = [container]
    pod_spec.restart_policy = "Never"
    return pod_template



from typing import Any

from fluid.utils.dispatcher import Dispatcher


class MessageDispatcher(Dispatcher[Any]):
    def event_type(self, data: Any) -> str:
        return type(data).__name__


dispatcher = MessageDispatcher()

assert dispatcher.dispatch({}) == 0


def count_keys(data: Any) -> None:
    print(f"number of keys {len(data)}")


dispatcher.register_handler("dict", count_keys)


assert dispatcher.dispatch(dict(a=1, b=2)) == 1



from enum import Enum
from typing import Any, Callable, Sequence, cast

from fastapi import APIRouter, Depends, FastAPI, HTTPException, Path, Query, Request
from pydantic import BaseModel, Field
from typing_extensions import Annotated, Doc

from fluid.tools_fastapi import app_workers
from fluid.utils.worker import Worker

from .broker import TaskManagersStatus
from .consumer import TaskManager
from .errors import UnknownTaskError
from .models import EmptyParams, Task, TaskInfo, TaskPriority, TaskRun


def get_task_manager_from_request(request: Request) -> TaskManager:
    return get_task_manager(request.app)


def get_task_manager(app: FastAPI) -> TaskManager:
    return cast(TaskManager, app.state.task_manager)


TaskManagerDep = Annotated[TaskManager, Depends(get_task_manager_from_request)]


class TaskUpdate(BaseModel):
    enabled: bool = Field(description="Task enabled or disabled")


class TaskCreate(BaseModel):
    name: str = Field(description="Task name")
    params: dict[str, Any] = Field(description="Task parameters", default_factory=dict)
    priority: TaskPriority | None = Field(default=None, description="Task priority")


def get_router(task_manager: TaskManager) -> APIRouter:
    router = APIRouter()

    router.add_api_route(
        "",
        get_tasks,
        methods=["GET"],
        response_model=list[TaskInfo],
        summary="List Tasks",
        description="Retrieve a list of tasks runs",
    )

    router.add_api_route(
        "-status",
        get_task_status,
        methods=["GET"],
        response_model=TaskManagersStatus,
        summary="Task managers status",
        description="Status of all running task managers",
    )

    router.add_api_route(
        "-queue",
        delete_queue,
        methods=["DELETE"],
        response_model=dict[str, int],
        summary="Clear all task queues",
        description="Clear all task queues, returns removed task count per priority",
    )

    router.add_api_route(
        "-queue/{priority}",
        delete_queue_by_priority,
        methods=["DELETE"],
        response_model=dict[str, int],
        summary="Clear a task queue",
        description="Clear a task queue by priority, returns number of removed tasks",
    )

    router.add_api_route(
        "/{task_name}",
        get_task,
        methods=["GET"],
        response_model=TaskInfo,
        summary="Get a Task",
        description="Retrieve information about a task",
    )

    router.add_api_route(
        "/{task_name}",
        patch_task,
        methods=["PATCH"],
        response_model=TaskInfo,
        summary="Update a task",
        description="Update a task configuration and enable/disable it",
    )

    for task in task_manager.registry.values():
        router.add_api_route(
            f"/{task.name}",
            (
                create_queue_task_no_params(task)
                if task.params_model is EmptyParams
                else create_queue_task(task)
            ),
            methods=["POST"],
            response_model=TaskRun,
            summary=f"Queue a new {task.name} task",
            description=f"Queue a new {task.name} task to be run",
        )

    return router


def create_queue_task_no_params(task: Task) -> Callable:

    async def queue_task(
        task_manager: TaskManagerDep,
    ) -> TaskRun:
        return await task_manager.queue(task)

    return queue_task


def create_queue_task(task: Task) -> Callable:
    TaskParams = task.params_model  # noqa

    async def queue_task(
        task_manager: TaskManagerDep,
        params: TaskParams,  # type: ignore [valid-type]
    ) -> TaskRun:
        kwargs: Any = params.model_dump() if params is not None else {}
        return await task_manager.queue(task, **kwargs)

    return queue_task


async def get_tasks(
    task_manager: TaskManagerDep,
    tags: Annotated[
        list[str] | None,
        Query(description="Only return tasks that have at least one of these tags"),
    ] = None,
) -> list[TaskInfo]:
    tasks = await task_manager.broker.get_tasks_info()
    if tags:
        wanted = set(tags)
        tasks = [task for task in tasks if wanted.intersection(task.tags)]
    return tasks


async def get_task(
    task_manager: TaskManagerDep,
    task_name: str = Path(title="Task name"),
) -> TaskInfo:
    data = await task_manager.broker.get_tasks_info(task_name)
    if not data:
        raise HTTPException(status_code=404, detail="Task not found")
    return data[0]


async def get_task_status(task_manager: TaskManagerDep) -> TaskManagersStatus:
    return await task_manager.broker.get_all_manager_statuses()


async def delete_queue(task_manager: TaskManagerDep) -> dict[str, int]:
    return await task_manager.broker.clear_queue()


async def delete_queue_by_priority(
    task_manager: TaskManagerDep,
    priority: Annotated[TaskPriority, Path(title="Queue priority")],
) -> dict[str, int]:
    return await task_manager.broker.clear_queue(priority)


async def patch_task(
    task_manager: TaskManagerDep,
    task_update: TaskUpdate,
    task_name: str = Path(title="Task name"),
) -> TaskInfo:
    try:
        return await task_manager.broker.enable_task(task_name, task_update.enabled)
    except UnknownTaskError as exc:
        raise HTTPException(status_code=404, detail="Task not found") from exc


def task_manager_fastapi(
    task_manager: Annotated[
        TaskManager,
        Doc(
            (
                "A [TaskManager][fluid.scheduler.TaskManager], "
                "[TaskConsumer][fluid.scheduler.TaskConsumer] or "
                "[TaskScheduler][fluid.scheduler.TaskScheduler] instance"
            )
        ),
    ],
    *,
    app: Annotated[
        FastAPI | None,
        Doc("FastAPI app instance. If not provided, a new instance is created."),
    ] = None,
    include_router: Annotated[
        bool,
        Doc("Whether to include the task manager router in the FastAPI app."),
    ] = True,
    prefix: Annotated[
        str,
        Doc("Prefix for the task manager routes."),
    ] = "/tasks",
    tags: Annotated[
        Sequence[str | Enum] | None,
        Doc("Tags for the task manager routes."),
    ] = None,
    **kwargs: Annotated[
        Any,
        Doc("Additional keyword arguments for the FastAPI app if not provided"),
    ],
) -> FastAPI:
    """Setup the FastAPI app and add the task manager to the state

    If the task manager is a [Worker][fluid.utils.worker.Worker], it is also added
    to the app workers to be started with the app.
    """
    app = app or FastAPI(**kwargs)
    if include_router:
        tags_ = tags if tags is not None else ["Tasks"]
        app.include_router(get_router(task_manager), prefix=prefix, tags=list(tags_))
    app.state.task_manager = task_manager
    if isinstance(task_manager, Worker):
        app_workers(app).add_workers(task_manager)
    else:
        app.router.on_startup.append(task_manager.on_startup)
        app.router.on_shutdown.append(task_manager.on_shutdown)
    return app



import asyncio
import json
import os
import time
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from enum import StrEnum, auto
from typing import Any, Self, Sequence, cast

from fastapi import FastAPI
from pydantic import BaseModel, Field

from fluid.scheduler import (
    RetryPolicy,
    TaskManagerPlugin,
    TaskRun,
    TaskScheduler,
    every,
    task,
    task_manager_fastapi,
)
from fluid.scheduler.broker import RedisTaskBroker
from fluid.utils.dates import utcnow
from fluid.utils.http_client import HttpxClient


@dataclass
class Deps:
    http_client: HttpxClient = field(default_factory=HttpxClient)

    @classmethod
    def get(cls, context: TaskRun) -> Self:
        return context.deps


def task_scheduler(
    *,
    deps: Deps | None = None,
    plugins: Sequence[TaskManagerPlugin] | None = None,
    **kwargs: Any,
) -> TaskScheduler:
    deps = deps or Deps()
    task_manager = TaskScheduler(deps=deps, **kwargs)
    task_manager.add_async_context_manager(deps.http_client)
    task_manager.register_from_dict(globals())
    for plugin in plugins or []:
        task_manager.with_plugin(plugin)
    return task_manager


def task_app() -> FastAPI:
    return task_manager_fastapi(task_scheduler(), app=FastAPI(title="Task Manager API"))


class Sleep(BaseModel):
    sleep: float = Field(default=0.1, ge=0, description="Sleep time")
    error: bool = False
    abort: bool = False


@task(
    max_concurrency=1,
    timeout_seconds=2,
    tags=("test", "fast"),
)
async def fast(context: TaskRun[Sleep]) -> None:
    """A task that sleeps for a while but has a 2 seconds timeout"""
    await asyncio.sleep(context.params.sleep)
    if context.params.error:
        raise RuntimeError("just an error")


@task(
    max_concurrency=1,
    timeout_seconds=120,
    tags=("test", "slow"),
)
async def dummy(context: TaskRun[Sleep]) -> None:
    """A task that sleeps for a while or errors"""
    await asyncio.sleep(context.params.sleep)
    if context.params.abort:
        context.abort("aborted by request")
    if context.params.error:
        raise RuntimeError("just an error")


@task(
    schedule=every(timedelta(seconds=2)),
    tags=("skip_db",),
)
async def ping(context: TaskRun) -> None:
    """A simple scheduled task that ping the broker"""
    redis_cli = cast(RedisTaskBroker, context.task_manager.broker).redis_cli
    await redis_cli.ping()


class AddValues(BaseModel):
    a: float = Field(default=0, description="First number to add")
    b: float = Field(default=0, description="Second number to add")


@task
async def add(context: TaskRun[AddValues]) -> None:
    """Log the addition of two numbers"""
    c = context.params.a + context.params.b
    context.logger.info(f"Adding {context.params.a} + {context.params.b} = {c}")


class DatetimeParams(BaseModel):
    dt: datetime = Field(default_factory=utcnow, description="A datetime parameter")


@task
async def datetime_task(context: TaskRun[DatetimeParams]) -> None:
    """A task with a datetime parameter"""
    context.logger.info(f"Received datetime: {context.params.dt}")


@task(cpu_bound=True, env={"FLUID_TEST_ENV": "hello_from_env"})
async def cpu_bound_env(context: TaskRun) -> None:
    """A CPU bound task that reads an env variable injected via the env argument"""
    broker = cast(RedisTaskBroker, context.task_manager.broker)
    await broker.redis_cli.setex(context.id, 10, os.environ.get("FLUID_TEST_ENV", ""))


@task(cpu_bound=True)
async def cpu_bound(context: TaskRun[Sleep]) -> None:
    """A CPU bound task running on subprocess

    CPU bound tasks are executed on a subprocess to avoid blocking the event loop.
    """
    time.sleep(context.params.sleep)
    if context.params.error:
        raise RuntimeError("deliberate cpu bound error")
    if context.params.abort:
        context.abort("aborted by request")
    broker = cast(RedisTaskBroker, context.task_manager.broker)
    redis = broker.redis_cli
    data = dict(
        pid=os.getpid(),
        sleep=context.params.sleep,
    )
    await redis.setex(context.id, 10, json.dumps(data))


class Scrape(BaseModel):
    url: str = Field(default="https://httpbin.org/get", description="URL to scrape")


@task
async def scrape(context: TaskRun[Scrape]) -> None:
    """Scrape a website"""
    deps = Deps.get(context)
    response = await deps.http_client.get(context.params.url, callback=True)
    text = await response.text()
    context.logger.info(text)


class FailCount(BaseModel):
    fail_times: int = Field(
        default=1, description="Fail this many times before succeeding"
    )


@task(retry=RetryPolicy(max_attempts=3, wait=0.0))
async def retryable(context: TaskRun[FailCount]) -> None:
    """A task that fails fail_times times before succeeding, used to test retry logic"""
    if context.retry_attempt < context.params.fail_times:
        raise RuntimeError(f"deliberate failure attempt {context.retry_attempt}")


@task(
    max_concurrency=1,
    timeout_seconds=10,
    rate_limit_retry=RetryPolicy(max_attempts=5, wait=0.0),
)
async def exclusive(context: TaskRun[Sleep]) -> None:
    """Task with max_concurrency=1 and rate_limit_retry for rate-limit retry tests."""
    await asyncio.sleep(context.params.sleep)


class Palette(StrEnum):
    RED = auto()
    GREEN = auto()
    BLUE = auto()


class PaletteParams(BaseModel):
    color: Palette = Palette.RED


class OptionalPaletteParams(BaseModel):
    color: Palette | None = None


class RequiredPaletteParams(BaseModel):
    color: Palette


@task
async def colorize(context: TaskRun[PaletteParams]) -> None:
    """A task with a StrEnum parameter"""
    broker = cast(RedisTaskBroker, context.task_manager.broker)
    await broker.redis_cli.setex(context.id, 10, context.params.color)
    context.logger.info(f"Color: {context.params.color}")


@task
async def colorize_optional(context: TaskRun[OptionalPaletteParams]) -> None:
    """A task with an optional StrEnum parameter"""
    if context.params.color is not None:
        broker = cast(RedisTaskBroker, context.task_manager.broker)
        await broker.redis_cli.setex(context.id, 10, context.params.color)
    context.logger.info(f"Color: {context.params.color}")


@task
async def colorize_required(context: TaskRun[RequiredPaletteParams]) -> None:
    """A task with a required StrEnum parameter"""
    broker = cast(RedisTaskBroker, context.task_manager.broker)
    await broker.redis_cli.setex(context.id, 10, context.params.color)
    context.logger.info(f"Color: {context.params.color}")



from datetime import timedelta
from time import monotonic

import pytest

from fluid.scheduler import (
    RetryPolicy,
    TaskPriority,
    TaskRun,
    TaskScheduler,
    TaskState,
    task,
)
from fluid.scheduler.consumer import InProcessTaskQueue
from fluid.scheduler.errors import TaskRunError
from fluid.scheduler.models import TaskDecoratorError
from fluid.utils.dates import utcnow
from fluid.utils.text import create_uid
from fluid.utils.waiter import wait_for

pytestmark = pytest.mark.asyncio(loop_scope="module")


# ---------------------------------------------------------------------------
# RetryPolicy unit tests
# ---------------------------------------------------------------------------


async def test_delay_fixed():
    policy = RetryPolicy(wait=5.0, backoff=1.0)
    assert policy.delay(1) == 5.0
    assert policy.delay(2) == 5.0
    assert policy.delay(10) == 5.0


async def test_delay_exponential():
    policy = RetryPolicy(wait=1.0, backoff=2.0)
    assert policy.delay(1) == 1.0
    assert policy.delay(2) == 2.0
    assert policy.delay(3) == 4.0


async def test_delay_capped_by_max_wait():
    policy = RetryPolicy(wait=10.0, backoff=2.0, max_wait=15.0)
    assert policy.delay(1) == 10.0
    assert policy.delay(2) == 15.0


async def test_matches_all_exceptions_when_empty():
    policy = RetryPolicy()
    assert policy.matches(RuntimeError("any"))
    assert policy.matches(ValueError("also"))


async def test_matches_specific_exception_type():
    policy = RetryPolicy(exceptions=(ValueError,))
    assert policy.matches(ValueError("yes"))
    assert not policy.matches(RuntimeError("no"))


async def test_matches_subclass():
    policy = RetryPolicy(exceptions=(Exception,))
    assert policy.matches(RuntimeError("subclass of Exception"))


# ---------------------------------------------------------------------------
# Helpers for TaskRun unit tests
# ---------------------------------------------------------------------------


def _make_task_run(t, task_manager: TaskScheduler, **kwargs) -> TaskRun:
    return TaskRun(
        id=create_uid(),
        task=t,
        priority=TaskPriority.medium,
        params=t.params_model(),
        task_manager=task_manager,
        **kwargs,
    )


# ---------------------------------------------------------------------------
# _maybe_failure_retry unit tests
# ---------------------------------------------------------------------------


@task(retry=RetryPolicy(max_attempts=2, wait=0.5, backoff=2.0))
async def _retry_task(ctx: TaskRun) -> None:
    pass


@task
async def _no_retry_task(ctx: TaskRun) -> None:
    pass


@task(retry=RetryPolicy(max_attempts=1, exceptions=(ValueError,)))
async def _typed_retry_task(ctx: TaskRun) -> None:
    pass


@task(retry=RetryPolicy(max_attempts=None, wait=0.0))
async def _unlimited_retry_task(ctx: TaskRun) -> None:
    pass


async def test_failure_retry_no_policy_returns_none(task_scheduler: TaskScheduler):
    task_run = _make_task_run(_no_retry_task, task_scheduler)
    assert task_run._maybe_failure_retry(RuntimeError("boom")) is None


async def test_failure_retry_returns_copy_on_first_attempt(
    task_scheduler: TaskScheduler,
):
    task_run = _make_task_run(_retry_task, task_scheduler)
    retry = task_run._maybe_failure_retry(RuntimeError("boom"))
    assert retry is not None
    assert retry.retry_attempt == 1
    assert retry.execute_after is not None
    assert retry.state == TaskState.init
    assert retry.queued is None
    assert retry.start is None
    assert retry.end is None


async def test_failure_retry_increments_attempt(task_scheduler: TaskScheduler):
    task_run = _make_task_run(_retry_task, task_scheduler, retry_attempt=1)
    retry = task_run._maybe_failure_retry(RuntimeError("boom"))
    assert retry is not None
    assert retry.retry_attempt == 2


async def test_failure_retry_exhausted_returns_none(task_scheduler: TaskScheduler):
    task_run = _make_task_run(_retry_task, task_scheduler, retry_attempt=2)
    assert task_run._maybe_failure_retry(RuntimeError("boom")) is None


async def test_failure_retry_unmatched_exception_returns_none(
    task_scheduler: TaskScheduler,
):
    task_run = _make_task_run(_typed_retry_task, task_scheduler)
    assert task_run._maybe_failure_retry(RuntimeError("wrong type")) is None


async def test_failure_retry_matched_exception_returns_copy(
    task_scheduler: TaskScheduler,
):
    task_run = _make_task_run(_typed_retry_task, task_scheduler)
    retry = task_run._maybe_failure_retry(ValueError("right type"))
    assert retry is not None
    assert retry.retry_attempt == 1


async def test_failure_retry_unlimited_attempts(task_scheduler: TaskScheduler):
    task_run = _make_task_run(_unlimited_retry_task, task_scheduler, retry_attempt=999)
    retry = task_run._maybe_failure_retry(RuntimeError("always retry"))
    assert retry is not None
    assert retry.retry_attempt == 1000


# ---------------------------------------------------------------------------
# _maybe_rate_limit_retry unit tests
# ---------------------------------------------------------------------------


@task(max_concurrency=1, rate_limit_retry=RetryPolicy(max_attempts=2, wait=0.5))
async def _rate_limit_retry_task(ctx: TaskRun) -> None:
    pass


@task(max_concurrency=1)
async def _rate_limit_no_retry_task(ctx: TaskRun) -> None:
    pass


async def test_rate_limit_retry_not_triggered_when_under_limit(
    task_scheduler: TaskScheduler,
):
    task_run = _make_task_run(_rate_limit_retry_task, task_scheduler)
    result = task_run._maybe_rate_limit_retry(current_runs=0)
    assert result is None
    assert task_run.state == TaskState.init


async def test_rate_limit_retry_returns_copy_when_at_limit(
    task_scheduler: TaskScheduler,
):
    task_run = _make_task_run(_rate_limit_retry_task, task_scheduler)
    retry = task_run._maybe_rate_limit_retry(current_runs=1)
    assert retry is not None
    assert retry.rate_limit_attempt == 1
    assert retry.execute_after is not None


async def test_rate_limit_retry_exhausted_sets_rate_limited_state(
    task_scheduler: TaskScheduler,
):
    task_run = _make_task_run(
        _rate_limit_retry_task, task_scheduler, rate_limit_attempt=2
    )
    result = task_run._maybe_rate_limit_retry(current_runs=1)
    assert result is None
    assert task_run.state == TaskState.rate_limited


async def test_rate_limit_no_retry_policy_sets_rate_limited_state(
    task_scheduler: TaskScheduler,
):
    task_run = _make_task_run(_rate_limit_no_retry_task, task_scheduler)
    result = task_run._maybe_rate_limit_retry(current_runs=1)
    assert result is None
    assert task_run.state == TaskState.rate_limited


# ---------------------------------------------------------------------------
# Integration tests
# ---------------------------------------------------------------------------


async def test_retry_succeeds_after_failure(task_scheduler: TaskScheduler) -> None:
    initial_info = await task_scheduler.broker.get_tasks_info("retryable")
    initial_end = initial_info[0].last_run_end

    await task_scheduler.queue("retryable", fail_times=1)

    async def retry_succeeded() -> bool:
        info = await task_scheduler.broker.get_tasks_info("retryable")
        ti = info[0]
        return ti.last_run_end != initial_end and ti.last_run_state == TaskState.success

    await wait_for(retry_succeeded, timeout=10)


async def test_retry_exhausted_ends_in_failure(task_scheduler: TaskScheduler) -> None:
    initial_info = await task_scheduler.broker.get_tasks_info("retryable")
    initial_end = initial_info[0].last_run_end

    await task_scheduler.queue("retryable", fail_times=4)

    async def retry_failed() -> bool:
        info = await task_scheduler.broker.get_tasks_info("retryable")
        ti = info[0]
        return ti.last_run_end != initial_end and ti.last_run_state == TaskState.failure

    await wait_for(retry_failed, timeout=10)


# ---------------------------------------------------------------------------
# TaskRun property tests
# ---------------------------------------------------------------------------


async def test_task_run_name_id(task_scheduler: TaskScheduler):
    task_run = _make_task_run(_retry_task, task_scheduler)
    assert task_run.name_id == f"{_retry_task.name}.{task_run.id}"


async def test_task_run_in_queue(task_scheduler: TaskScheduler):
    task_run = _make_task_run(_retry_task, task_scheduler)
    assert task_run.in_queue is None
    task_run.set_state(TaskState.running)
    assert task_run.in_queue is not None


async def test_task_run_duration_and_total(task_scheduler: TaskScheduler):
    task_run = _make_task_run(_retry_task, task_scheduler)
    assert task_run.duration is None
    assert task_run.duration_ms is None
    assert task_run.total is None
    task_run.set_state(TaskState.success)
    assert task_run.duration is not None
    assert task_run.duration_ms is not None
    assert task_run.total is not None


async def test_set_state_same_state_is_noop(task_scheduler: TaskScheduler):
    task_run = _make_task_run(_retry_task, task_scheduler)
    task_run.set_state(TaskState.init)
    assert task_run.state == TaskState.init


async def test_set_state_invalid_transition_raises(task_scheduler: TaskScheduler):
    task_run = _make_task_run(_retry_task, task_scheduler)
    task_run.set_state(TaskState.success)
    with pytest.raises(TaskRunError):
        task_run.set_state(TaskState.running)


# ---------------------------------------------------------------------------
# task() decorator error cases
# ---------------------------------------------------------------------------


async def test_task_decorator_error_positional_and_kwargs():
    async def my_func(ctx: TaskRun) -> None:
        pass

    with pytest.raises(TaskDecoratorError):
        task(my_func, name="foo")  # type: ignore[call-overload]


async def test_task_decorator_error_no_executor_no_kwargs():
    with pytest.raises(TaskDecoratorError):
        task()


# ---------------------------------------------------------------------------
# rate_limit_retry integration test (covers consumer._register_task_run 517-518)
# ---------------------------------------------------------------------------


async def test_rate_limit_retry_requeues_and_succeeds(
    task_scheduler: TaskScheduler,
) -> None:
    initial_info = await task_scheduler.broker.get_tasks_info("exclusive")
    initial_end = initial_info[0].last_run_end

    # queue two concurrent runs — the second will be rate-limited and retried
    await task_scheduler.queue("exclusive", sleep=0.5)
    await task_scheduler.queue("exclusive", sleep=0.1)

    async def both_done() -> bool:
        info = await task_scheduler.broker.get_tasks_info("exclusive")
        ti = info[0]
        return ti.last_run_end != initial_end and ti.last_run_state == TaskState.success

    await wait_for(both_done, timeout=10)


# ---------------------------------------------------------------------------
# _re_queue_if_not_ready unit tests
# ---------------------------------------------------------------------------


async def test_re_queue_no_execute_after(task_scheduler: TaskScheduler) -> None:
    task_run = _make_task_run(_retry_task, task_scheduler)
    assert task_run.execute_after is None
    assert task_scheduler._re_queue_if_not_ready(task_run) is False


async def test_re_queue_execute_after_in_past(task_scheduler: TaskScheduler) -> None:
    task_run = _make_task_run(
        _retry_task,
        task_scheduler,
        execute_after=utcnow() - timedelta(seconds=10),
    )
    assert task_scheduler._re_queue_if_not_ready(task_run) is False


async def test_re_queue_execute_after_in_future(task_scheduler: TaskScheduler) -> None:
    task_run = _make_task_run(
        _retry_task,
        task_scheduler,
        execute_after=utcnow() + timedelta(seconds=60),
    )
    before = monotonic()
    result = task_scheduler._re_queue_if_not_ready(task_run)
    assert result is True
    entries = [e for e in task_scheduler._in_process_queue._heap if e[2] is task_run]
    assert len(entries) == 1
    assert entries[0][0] >= before + 5


async def test_re_queue_minimum_delay_is_5_seconds(
    task_scheduler: TaskScheduler,
) -> None:
    # execute_after only 1 s away — delay must be clamped to the 5 s minimum
    task_run = _make_task_run(
        _retry_task,
        task_scheduler,
        execute_after=utcnow() + timedelta(seconds=1),
    )
    before = monotonic()
    task_scheduler._re_queue_if_not_ready(task_run)
    entries = [e for e in task_scheduler._in_process_queue._heap if e[2] is task_run]
    assert len(entries) == 1
    assert entries[0][0] >= before + 5
    assert entries[0][0] < before + 6


async def test_in_process_queue_drains_on_stop(
    task_scheduler: TaskScheduler,
) -> None:
    task_run1 = _make_task_run(_retry_task, task_scheduler)
    task_run2 = _make_task_run(_retry_task, task_scheduler)

    queue = InProcessTaskQueue(task_scheduler)
    queue.queue(task_run1, delay=100)
    queue.queue(task_run2, delay=100)
    assert queue.queue_len() == 2

    # worker never started so is_running() is False — drain runs immediately
    await queue.run()

    assert queue.queue_len() == 0



from datetime import timedelta

import pytest

from fluid.db import CrudDB, Pagination
from fluid.utils.dates import isoformat, utcnow

pytestmark = pytest.mark.asyncio(loop_scope="module")


async def test_paginate(db: CrudDB) -> None:
    created = utcnow()
    dates = [created - timedelta(days=10 - i) for i in range(1, 11)]
    data = [
        dict(title=f"Example{i}", created=created - timedelta(days=10 - i))
        for i in range(1, 11)
    ]
    result = await db.db_insert(db.tables["tasks"], data)
    rows = result.all()
    assert len(rows) == 10
    p = Pagination.create(
        "created",
        filters={"created:ge": dates[1] + timedelta(seconds=30)},
        limit=5,
        desc=True,
    )
    rows, cursor = await p.execute(db, db.tables["tasks"])
    assert len(rows) == 5
    assert cursor
    p = Pagination.create(
        "created",
        cursor=cursor,
        desc=True,
    )
    extra_rows, cursor = await p.execute(db, db.tables["tasks"])
    assert len(extra_rows) == 3
    all_rows = list(rows)
    all_rows.extend(extra_rows)
    assert not cursor
    assert [isoformat(r.created) for r in all_rows] == [
        isoformat(d) for d in reversed(dates[2:])
    ]



from __future__ import annotations

import asyncio
from abc import ABC, abstractmethod
from collections import defaultdict
from typing import Annotated, Awaitable, Callable, Generic, NamedTuple, Self, TypeVar

from typing_extensions import Doc

MessageType = TypeVar("MessageType")
MessageHandlerType = TypeVar("MessageHandlerType")


class Event(NamedTuple):
    type: Annotated[str, Doc("The event type")]
    tag: Annotated[
        str,
        Doc("The event tag - for registering multiple handlers for a given event type"),
    ] = ""

    @classmethod
    def from_string_or_event(
        cls,
        event: Annotated[
            str | Self,
            Doc(
                "The [Event][fluid.utils.dispatcher.Event] or a string of"
                " the form `{event_type}` or `{event_type}.{event_tag}`"
            ),
        ],
    ) -> Self:
        if isinstance(event, str):
            return cls.from_string(event)
        return event

    @classmethod
    def from_string(
        cls,
        event: Annotated[
            str,
            Doc(
                "The event string has the form `{event_type}` "
                "or `{event_type}.{event_tag}`"
            ),
        ],
    ) -> Self:
        bits = event.split(".")
        return cls(bits[0], bits[1] if len(bits) > 1 else "")


class BaseDispatcher(Generic[MessageType, MessageHandlerType], ABC):
    """Base generic abstract class for dispatchers"""

    def __init__(self) -> None:
        self._msg_handlers: defaultdict[str, dict[str, MessageHandlerType]] = (
            defaultdict(
                dict,
            )
        )

    def register_handler(
        self,
        event: Annotated[Event | str, Doc("The event to register the handler for")],
        handler: Annotated[MessageHandlerType, Doc("The handler to register")],
    ) -> MessageHandlerType | None:
        """Register a handler for the given event

        It is possible to register multiple handlers for the same event type by
        providing a different tag for each handler.

        For example, to register two handlers for the event type `foo`:

        ```python
        dispatcher.register_handler("foo.first", handler1)
        dispatcher.register_handler("foo.second", handler2)
        ```
        """
        event = Event.from_string_or_event(event)
        previous = self._msg_handlers[event.type].get(event.tag)
        self._msg_handlers[event.type][event.tag] = handler
        return previous

    def unregister_handler(
        self, event: Annotated[Event | str, Doc("The event to unregister the handler")]
    ) -> MessageHandlerType | None:
        """Unregister a handler for the given event

        It returns the handler that was unregistered or `None` if no handler was
        registered for the given event.
        """
        event = Event.from_string_or_event(event)
        return self._msg_handlers[event.type].pop(event.tag, None)

    def get_handlers(
        self,
        message: Annotated[MessageType, Doc("The message to get the handlers for")],
    ) -> dict[str, MessageHandlerType] | None:
        """Get all event handlers for the given message

        This method returns a dictionary of all handlers registered for the given
        message type. If no handlers are registered for the message type, it returns
        `None`.
        """
        event_type = self.event_type(message)
        return self._msg_handlers.get(event_type)

    @abstractmethod
    def event_type(self, message: MessageType) -> str:
        """return the event type as string"""


class Dispatcher(BaseDispatcher[MessageType, Callable[[MessageType], None]]):
    """Dispatcher for sync handlers"""

    def dispatch(self, message: MessageType) -> int:
        """dispatch the message to all handlers

        It returns the number of handlers that were called
        """
        handlers = self.get_handlers(message)
        if handlers:
            for handler in handlers.values():
                handler(message)
        return len(handlers or ())


class AsyncDispatcher(
    BaseDispatcher[MessageType, Callable[[MessageType], Awaitable[None]]],
):
    """Dispatcher for async handlers"""

    async def dispatch(self, message: MessageType) -> int:
        """Dispatch the message and wait for all handlers to complete

        It returns the number of handlers that were called
        """
        handlers = self.get_handlers(message)
        if handlers:
            await asyncio.gather(*[handler(message) for handler in handlers.values()])
        return len(handlers or ())



"""initial

Revision ID: b2ee259b4966
Revises:
Create Date: 2025-07-07 15:45:05.882984

"""

from typing import Sequence, Union

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "b2ee259b4966"
down_revision: Union[str, None] = None
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "multi_key",
        sa.Column("x", sa.JSON(), nullable=True),
        sa.Column("y", sa.JSON(), nullable=True),
    )
    op.create_table(
        "multi_key_unique",
        sa.Column("x", sa.Integer(), nullable=False),
        sa.Column("y", sa.Integer(), nullable=False),
        sa.UniqueConstraint("x", "y"),
    )
    op.create_table(
        "tasks",
        sa.Column("id", sa.String(length=32), nullable=False),
        sa.Column("title", sa.String(length=64), nullable=False),
        sa.Column("created", sa.DateTime(timezone=True), nullable=True),
        sa.Column("done", sa.DateTime(timezone=True), nullable=True),
        sa.Column("severity", sa.Integer(), nullable=True),
        sa.Column("created_by", sa.String(), nullable=False),
        sa.Column("type", sa.Enum("todo", "issue", name="tasktype"), nullable=True),
        sa.Column("unique_title", sa.String(), nullable=True),
        sa.Column("story_points", sa.Numeric(), nullable=True),
        sa.Column("random", sa.String(length=64), nullable=True),
        sa.Column("subtitle", sa.String(length=64), nullable=False),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("unique_title"),
    )
    op.create_table(
        "randoms",
        sa.Column("id", sa.String(length=32), nullable=False),
        sa.Column("randomdate", sa.Date(), nullable=False),
        sa.Column("timestamp", sa.DateTime(timezone=True), nullable=False),
        sa.Column("price", sa.Numeric(precision=100, scale=4), nullable=False),
        sa.Column("tenor", sa.String(length=3), nullable=False),
        sa.Column("tick", sa.Boolean(), nullable=True),
        sa.Column("info", sa.JSON(), nullable=True),
        sa.Column("jsonlist", sa.JSON(), nullable=True),
        sa.Column("task_id", sa.String(length=32), nullable=False),
        sa.ForeignKeyConstraint(["task_id"], ["tasks.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("id"),
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("randoms")
    op.drop_table("tasks")
    op.drop_table("multi_key_unique")
    op.drop_table("multi_key")
    # ### end Alembic commands ###



from pathlib import Path

from fluid.db import CrudDB

from .tables1 import meta
from .tables2 import additional_meta

MIGRATIONS_PATH = Path(__file__).parent / "migrations"


def setup_tables(db: CrudDB) -> CrudDB:
    meta(db.metadata)
    additional_meta(db.metadata)
    return db


def get_db() -> CrudDB:
    return setup_tables(CrudDB.from_env(migration_path=MIGRATIONS_PATH))



from __future__ import annotations

import asyncio
import enum
import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import (
    Annotated,
    Awaitable,
    Callable,
    Iterator,
    Protocol,
    Self,
    Sequence,
    TypeVar,
)

from inflection import underscore
from typing_extensions import Doc

from fluid import settings

from .dispatcher import AsyncDispatcher, MessageType
from .errors import WorkerStartError

logger = logging.getLogger(__name__)


class WorkerState(enum.StrEnum):
    """The lifecycle state of a [Worker][fluid.utils.worker.Worker]."""

    INIT = enum.auto()
    """Worker has been created but not yet started."""
    RUNNING = enum.auto()
    """Worker is actively executing its [run][fluid.utils.worker.Worker.run] loop."""
    STOPPING = enum.auto()
    """Graceful stop requested; [run][fluid.utils.worker.Worker.run] should
    exit at its next safe point."""
    STOPPED = enum.auto()
    """Worker exited cleanly after a graceful stop."""
    FORCE_STOPPED = enum.auto()
    """Worker was cancelled because it did not exit within the grace period."""


class HasWorkers(ABC):
    @abstractmethod
    def get_workers(self) -> Sequence[Worker]:
        """Get the workers that can be added to a the Workers Manager"""


@dataclass
class WorkerTaskRunner:
    worker: Worker
    task: asyncio.Task | None = None
    started_shutdown: bool = False
    force_shutdown: bool = False

    @classmethod
    async def start(cls, worker: Worker) -> Self:
        worker_task = cls(worker)
        waiter = asyncio.Future[Self]()
        worker_task.task = asyncio.create_task(
            worker_task.run(waiter), name=worker.worker_name
        )
        return await waiter

    async def shutdown(self) -> None:
        if self.task is None:  # pragma: no cover
            return
        elif not self.started_shutdown:
            self.started_shutdown = True
            await self.gracefully_shutdown()
        await self.wait_for_shutdown()

    async def wait_for_shutdown(self) -> None:
        if self.task is None:
            return
        await self.task

    async def gracefully_shutdown(self) -> None:
        self.started_shutdown = True
        self.worker.gracefully_stop()
        try:
            async with asyncio.timeout(self.worker._stopping_grace_period):
                while self.task is not None:
                    await asyncio.sleep(0)
            return None
        except asyncio.TimeoutError:
            pass
        if self.task is None:  # pragma: no cover
            return None
        self.force_shutdown = True
        self.task.cancel()
        logger.warning("%s forced shutdown", self.worker.worker_name)

    async def run(self, waiter: asyncio.Future) -> None:
        if self.worker.is_running():
            raise WorkerStartError("worker %s already running", self.worker.worker_name)
        self.worker._worker_state = WorkerState.RUNNING
        logger.info("%s started running", self.worker.worker_name)
        waiter.set_result(self)
        exit_reason = "normal shutdown"
        exit_code = 0
        stopped_state = WorkerState.STOPPED
        try:
            await self.worker.on_startup()
            await self.worker.run()
            await self.worker.on_shutdown()
        except Exception as e:
            exit_reason = str(e)
            exit_code = 2
            raise
        except asyncio.CancelledError as e:
            if self.force_shutdown:
                exit_reason = "forced shutdown"
                exit_code = 1
                stopped_state = WorkerState.FORCE_STOPPED
                # we are shutting down, this is expected
                pass
            else:
                exit_reason = str(e)
                exit_code = 2
                raise
        finally:
            logger.warning(
                "%s stopped running: %s",
                self.worker.worker_name,
                exit_reason,
            )
            asyncio.get_event_loop().call_soon(
                self.worker.after_shutdown,
                exit_reason,
                exit_code,
            )
            self.worker._worker_state = stopped_state
            self.worker._worker_task_runner = None
            self.task = None


class Worker(ABC):
    """Abstract base class for all workers.

    A worker encapsulates a long-running async task with a managed lifecycle.
    Subclasses implement [run][fluid.utils.worker.Worker.run], which is called
    once the worker is started and should loop until
    [is_running][fluid.utils.worker.Worker.is_running] returns `False`.

    Use [startup][fluid.utils.worker.Worker.startup] to start the worker as an
    asyncio task, and [shutdown][fluid.utils.worker.Worker.shutdown] (or
    [gracefully_stop][fluid.utils.worker.Worker.gracefully_stop] +
    [wait_for_shutdown][fluid.utils.worker.Worker.wait_for_shutdown]) to stop it.

    Override [on_startup][fluid.utils.worker.Worker.on_startup] and
    [on_shutdown][fluid.utils.worker.Worker.on_shutdown] to initialise and clean
    up async resources that the worker owns.
    """

    def __init__(
        self,
        *,
        name: Annotated[
            str,
            Doc("Worker's name, if not provided it is evaluated from the class name"),
        ] = "",
        stopping_grace_period: Annotated[
            float,
            Doc(
                "Grace period in seconds to wait for workers to stop running "
                "when this worker is shutdown. "
                "It defaults to the `FLUID_STOPPING_GRACE_PERIOD` "
                "environment variable or 10 seconds."
            ),
        ] = settings.STOPPING_GRACE_PERIOD,
    ) -> None:
        self._worker_name: str = name or underscore(type(self).__name__)
        self._worker_state: WorkerState = WorkerState.INIT
        self._stopping_grace_period = stopping_grace_period
        self._worker_task_runner: WorkerTaskRunner | None = None

    @property
    def worker_state(self) -> WorkerState:
        """The running state of the worker"""
        return self._worker_state

    @property
    def worker_name(self) -> str:
        """The name of the worker"""
        return self._worker_name

    @property
    def num_workers(self) -> int:
        """The number of workers in this worker"""
        return 1

    def has_started(self) -> bool:
        return self._worker_state != WorkerState.INIT

    def is_running(self) -> bool:
        return self._worker_state == WorkerState.RUNNING

    def is_stopping(self) -> bool:
        return self._worker_state == WorkerState.STOPPING

    def is_stopped(self) -> bool:
        return self._worker_state in (WorkerState.STOPPED, WorkerState.FORCE_STOPPED)

    def gracefully_stop(self) -> None:
        """Try to gracefully stop the worker"""
        if self.is_running():
            self._worker_state = WorkerState.STOPPING

    def after_shutdown(self, reason: str, code: int) -> None:  # noqa: B027
        """Called after shutdown of worker

        By default it does nothing, but can be overriden to do something such as
        exit the process.
        """

    async def status(self) -> dict:
        return {"stopping": self.is_stopping(), "running": self.is_running()}

    async def on_startup(self) -> None:  # noqa: B027
        """Called when the worker starts running

        Use this function to initialize other async resources connected with the worker
        """

    async def on_shutdown(self) -> None:  # noqa: B027
        """called after the worker stopped running

        Use this function to cleanup resources connected with the worker
        """

    async def __aenter__(self) -> Self:
        """Start the worker, allowing it to be used as an async context manager."""
        await self.startup()
        return self

    async def __aexit__(self, *exc_info: object) -> None:
        """Shut down the worker."""
        await self.shutdown()

    async def startup(self) -> None:
        """start the worker

        This method creates a task to run the worker.
        """
        if self.has_started():
            raise WorkerStartError(
                "worker %s already started: %s", self.worker_name, self._worker_state
            )
        else:
            self._worker_task_runner = await WorkerTaskRunner.start(self)

    async def shutdown(self) -> None:
        """Shutdown a running worker and wait for it to stop

        This method will try to gracefully stop the worker and wait for it to stop.
        If the worker does not stop in the grace period, it will force shutdown
        by cancelling the task.
        """
        if self._worker_task_runner is not None:
            await self._worker_task_runner.shutdown()

    async def wait_for_shutdown(self) -> None:
        """Wait for the worker to stop

        This method will wait for the worker to stop running, but doesn't
        try to gracefully stop it nor force shutdown.
        """
        if self._worker_task_runner is not None:
            await self._worker_task_runner.wait_for_shutdown()

    def workers(self) -> Iterator[Worker]:
        """An iterator of workers in this worker"""
        yield self

    @abstractmethod
    async def run(self) -> None:
        """run the worker

        This is the only abstract method and that needs implementing.
        It is the coroutine that mantains the worker running.
        """


class WorkerFunction(Worker):
    """A [Worker][fluid.utils.worker.Worker] that calls a coroutine function in a loop.

    On each iteration the supplied `run_function` is awaited, then the worker
    sleeps for `heartbeat` seconds before repeating. The loop exits when
    [is_running][fluid.utils.worker.Worker.is_running] returns `False`.
    """

    def __init__(
        self,
        run_function: Annotated[
            Callable[[], Awaitable[None]],
            Doc(
                "The coroutine function tuo run and await at each iteration "
                "of the worker loop"
            ),
        ],
        *,
        heartbeat: Annotated[
            float | int, Doc("The time to wait between each coroutine function run")
        ] = 0,
        name: Annotated[
            str,
            Doc("Worker's name, if not provided it is evaluated from the class name"),
        ] = "",
        stopping_grace_period: Annotated[
            float,
            Doc("Grace period in seconds before force-cancelling this worker"),
        ] = settings.STOPPING_GRACE_PERIOD,
    ) -> None:
        super().__init__(name=name, stopping_grace_period=stopping_grace_period)
        self._run_function = run_function
        self._heartbeat = heartbeat

    async def run(self) -> None:
        while self.is_running():
            await self._run_function()
            await asyncio.sleep(self._heartbeat)


T = TypeVar("T", contravariant=True)


class MessageProducer(Protocol[T]):
    def send(self, message: T | None) -> None: ...


class QueueConsumer(Worker, MessageProducer[MessageType]):
    """Abstract [Worker][fluid.utils.worker.Worker] backed by an asyncio queue.

    Provides [send][fluid.utils.worker.QueueConsumer.send] for thread-safe
    message delivery and [get_message][fluid.utils.worker.QueueConsumer.get_message]
    for retrieving the next message with a timeout. Subclasses implement
    [run][fluid.utils.worker.Worker.run] to consume messages from the queue.
    """

    def __init__(
        self,
        *,
        name: Annotated[
            str,
            Doc("Worker's name, if not provided it is evaluated from the class name"),
        ] = "",
        stopping_grace_period: Annotated[
            float,
            Doc(
                "Grace period in seconds to wait for workers to stop running "
                "when this worker is shutdown. "
                "It defaults to the `FLUID_STOPPING_GRACE_PERIOD` "
                "environment variable or 10 seconds."
            ),
        ] = settings.STOPPING_GRACE_PERIOD,
    ) -> None:
        super().__init__(name=name, stopping_grace_period=stopping_grace_period)
        self._queue: asyncio.Queue[MessageType | None] = asyncio.Queue()

    async def get_message(self, timeout: float = 0.5) -> MessageType | None:
        """Get the next message from the queue"""
        try:
            async with asyncio.timeout(timeout):
                return await self._queue.get()
        except asyncio.TimeoutError:
            return None
        except (asyncio.CancelledError, RuntimeError):
            if not self.is_stopping():
                raise
        return None

    def queue_size(self) -> int:
        """Get the size of the queue"""
        return self._queue.qsize()

    async def status(self) -> dict:
        status = await super().status()
        status.update(queue_size=self.queue_size())
        return status

    def send(self, message: MessageType | None) -> None:
        """Send a message into the worker"""
        self._queue.put_nowait(message)


class QueueConsumerWorker(QueueConsumer[MessageType]):
    """A [QueueConsumer][fluid.utils.worker.QueueConsumer] that dispatches each
    message to a single async callback."""

    def __init__(
        self,
        on_message: Annotated[
            Callable[[MessageType], Awaitable[None]],
            Doc("The async callback to call when a message is received"),
        ],
        *,
        name: Annotated[
            str,
            Doc("Worker's name, if not provided it is evaluated from the class name"),
        ] = "",
        stopping_grace_period: Annotated[
            float,
            Doc(
                "Grace period in seconds to wait for workers to stop running "
                "when this worker is shutdown. "
                "It defaults to the `FLUID_STOPPING_GRACE_PERIOD` "
                "environment variable or 10 seconds."
            ),
        ] = settings.STOPPING_GRACE_PERIOD,
    ) -> None:
        super().__init__(name=name, stopping_grace_period=stopping_grace_period)
        self.on_message = on_message

    async def run(self) -> None:
        while not self.is_stopping():
            message = await self.get_message()
            if message is not None:
                await self.on_message(message)


class AsyncConsumer(QueueConsumer[MessageType]):
    """A [QueueConsumer][fluid.utils.worker.QueueConsumer] that fans out each
    message to all registered async handlers via an
    [AsyncDispatcher][fluid.utils.dispatcher.AsyncDispatcher].

    The run loop processes messages until
    [is_stopping][fluid.utils.worker.Worker.is_stopping] returns `True`.
    Any messages remaining in the queue when the worker stops are discarded;
    callers that need guaranteed delivery should drain the queue before
    requesting a stop.
    """

    def __init__(
        self,
        dispatcher: Annotated[
            AsyncDispatcher[MessageType],
            Doc("Async message dispatcher to dispatch messages"),
        ],
        *,
        name: Annotated[
            str,
            Doc("Worker's name, if not provided it is evaluated from the class name"),
        ] = "",
        stopping_grace_period: Annotated[
            float,
            Doc(
                "Grace period in seconds to wait for workers to stop running "
                "when this worker is shutdown. "
                "It defaults to the `FLUID_STOPPING_GRACE_PERIOD` "
                "environment variable or 10 seconds."
            ),
        ] = settings.STOPPING_GRACE_PERIOD,
    ) -> None:
        super().__init__(name=name, stopping_grace_period=stopping_grace_period)
        self.dispatcher: AsyncDispatcher[MessageType] = dispatcher

    async def run(self) -> None:
        while not self.is_stopping():
            message = await self.get_message()
            if message is not None:
                await self.dispatcher.dispatch(message)


class Workers(Worker):
    """A [Worker][fluid.utils.worker.Worker] that owns and manages a collection
    of child workers.

    Child workers are registered with
    [add_workers][fluid.utils.worker.Workers.add_workers].
    When the `Workers` instance starts, its [run][fluid.utils.worker.Workers.run] loop
    starts each child worker and monitors their health — if any child stops
    unexpectedly the whole group is gracefully stopped.

    On shutdown all child workers are stopped concurrently via
    [_wait_for_workers][fluid.utils.worker.Workers._wait_for_workers].
    Workers that do not exit within `stopping_grace_period` seconds are
    force-cancelled.

    !!! note "Shutdown ordering"
        By default all child workers are shut down concurrently. If a subclass
        needs a specific shutdown order (e.g. an event dispatcher that must
        outlive the workers that produce events), override
        [_wait_for_workers][fluid.utils.worker.Workers._wait_for_workers].
    """

    def __init__(
        self,
        *workers: Annotated[
            Worker,
            Doc(
                "Workers to manage, they can also be added later "
                "via `add_workers` method"
            ),
        ],
        name: Annotated[
            str,
            Doc("Worker's name, if not provided it is evaluated from the class name"),
        ] = "",
        heartbeat: Annotated[
            float | int,
            Doc("The time to wait between each workers status check"),
        ] = 0.1,
        stopping_grace_period: Annotated[
            float,
            Doc(
                "Grace period in seconds to wait for workers to stop running "
                "when this worker is shutdown. "
                "It defaults to the `FLUID_STOPPING_GRACE_PERIOD` "
                "environment variable or 10 seconds."
            ),
        ] = settings.STOPPING_GRACE_PERIOD,
    ) -> None:
        super().__init__(name=name, stopping_grace_period=stopping_grace_period)
        self._heartbeat = heartbeat
        self._workers: list[Worker] = []
        self.add_workers(*workers)

    def add_workers(self, *workers: Worker) -> None:
        """add workers to the workers

        They can be added while the worker is running.
        """
        for worker in workers:
            if worker not in self._workers:
                self._workers.append(worker)

    @property
    def num_workers(self) -> int:
        return len(self._workers)

    def workers(self) -> Iterator[Worker]:
        return iter(self._workers)

    def gracefully_stop(self) -> None:
        """Try to gracefully stop the workers and this worker"""
        super().gracefully_stop()
        for worker in self._workers:
            worker.gracefully_stop()

    async def status(self) -> dict:
        status_workers = await asyncio.gather(
            *[worker.status() for worker in self._workers],
        )
        return {
            worker.worker_name: status
            for worker, status in zip(self._workers, status_workers, strict=False)
        }

    async def run(self) -> None:
        while self.is_running():
            for worker in self._workers:
                if not worker.has_started():
                    await worker.startup()
                if not worker.is_running():
                    self.gracefully_stop()
                    break
            await asyncio.sleep(self._heartbeat)
        await self._wait_for_workers()

    async def _wait_for_workers(self) -> None:
        await asyncio.gather(
            *[worker.shutdown() for worker in self._workers], return_exceptions=True
        )



from uuid import uuid4

from fluid.utils.text import as_uuid


def test_as_uuid() -> None:
    assert as_uuid(None) is None
    assert as_uuid("a") is None
    uid = uuid4()
    assert as_uuid(uid) == uid.hex
    assert as_uuid(uid.hex) == uid.hex



from typing import Iterator

import pytest
from fastapi import FastAPI
from httpx import ASGITransport, AsyncClient
from redis.asyncio import Redis

from examples import tasks
from examples.db import get_db
from fluid.scheduler import TaskManager, TaskScheduler
from fluid.scheduler.db import TaskDbPlugin
from fluid.scheduler.endpoints import get_task_manager, task_manager_fastapi
from fluid.utils.stacksampler import Sampler
from tests.scheduler.tasks import TaskClient, redis_broker, start_fastapi


@pytest.fixture(scope="module")
def db_plugin() -> TaskDbPlugin:
    db = get_db()
    plugin = TaskDbPlugin(db)
    mig = db.migration()
    if not mig.db_create():
        mig.drop_all_schemas()
    mig.create_all()
    return plugin


@pytest.fixture(scope="module", autouse=True)
def sampler() -> Iterator[Sampler]:
    sampler = Sampler()
    sampler.start()
    yield sampler
    sampler.stop()


@pytest.fixture(scope="module")
async def task_app():
    task_manager = tasks.task_scheduler(max_concurrent_tasks=2, schedule_tasks=True)
    broker = redis_broker(task_manager)
    await broker.clear()
    async with start_fastapi(task_manager_fastapi(task_manager)) as app:
        yield app


@pytest.fixture(scope="module")
async def task_scheduler(task_app: FastAPI) -> TaskManager:
    return get_task_manager(task_app)


@pytest.fixture(scope="module")
def redis(task_scheduler: TaskScheduler) -> Redis:
    return redis_broker(task_scheduler).redis_cli


@pytest.fixture(scope="module")
async def cli(task_app):
    base_url = TaskClient().url
    async with AsyncClient(
        transport=ASGITransport(app=task_app), base_url=base_url
    ) as session:
        async with TaskClient(url=base_url, session=session) as client:
            yield client



from datetime import datetime

import pytest

from fluid.db import CrudDB

pytestmark = pytest.mark.asyncio(loop_scope="module")


async def test_upsert(db: CrudDB) -> None:
    task = await db.db_upsert(
        db.tables["tasks"], dict(title="Example"), dict(severity=4)
    )
    assert task.id
    assert task.severity == 4
    assert task.done is None
    task2 = await db.db_upsert(
        db.tables["tasks"], dict(title="Example"), dict(done=datetime.now())
    )
    assert task2.id == task.id
    assert task2.done
    assert await db.db_count(db.tables["tasks"], {}) == 1


async def test_upsert_no_data(db: CrudDB) -> None:
    task = await db.db_upsert(db.tables["tasks"], dict(title="Example2"))
    assert task.id
    assert task.title == "Example2"


async def test_insert_missing_columns(db: CrudDB) -> None:
    table = db.tables["tasks"]
    result = await db.db_insert(
        table,
        [dict(title="Task1", severity=1), dict(title="Task2")],
    )
    rows = result.fetchall()
    assert len(rows) == 2
    titles = {r.title for r in rows}
    assert titles == {"Task1", "Task2"}
    severities = {r.title: r.severity for r in rows}
    assert severities["Task1"] == 1
    assert severities["Task2"] is None


async def test_search_query(db: CrudDB) -> None:
    table = db.tables["tasks"]
    await db.db_insert(table, dict(title="SearchMe", random="foo"))
    await db.db_insert(table, dict(title="IgnoreMe", random="bar"))
    # match on title
    sql = db.search_query(table, table.select(), ("title",), "SearchMe")
    async with db.ensure_connection() as conn:
        rows = (await conn.execute(sql)).fetchall()
    assert all("SearchMe" in r.title for r in rows)
    assert not any("IgnoreMe" in r.title for r in rows)
    # no-op when search string is empty
    sql_noop = db.search_query(table, table.select(), ("title",), "")
    async with db.ensure_connection() as conn:
        all_rows = (await conn.execute(sql_noop)).fetchall()
    assert len(all_rows) >= 2
    # no-op when search_fields is empty
    sql_noop2 = db.search_query(table, table.select(), (), "SearchMe")
    async with db.ensure_connection() as conn:
        all_rows2 = (await conn.execute(sql_noop2)).fetchall()
    assert len(all_rows2) >= 2


async def test_delete(db: CrudDB) -> None:
    table = db.tables["tasks"]
    task = await db.db_upsert(table, dict(title="ToDelete"))
    assert await db.db_count(table, dict(title="ToDelete")) == 1
    result = await db.db_delete(table, dict(title="ToDelete"))
    deleted = result.fetchall()
    assert len(deleted) == 1
    assert deleted[0].id == task.id
    assert await db.db_count(table, dict(title="ToDelete")) == 0
